"""Use VidStab as command line script

usage: python -m vidstab [-h] -i INPUT -o OUTPUT [-p PLAYBACK] [-k KEYPOINTMETHOD]
                   [-s SMOOTHWINDOW] [-m MAXFRAMES] [-b BORDERTYPE]
                   [-z BORDERSIZE] [-l LAYERFRAMES]

optional arguments:
  -h, --help            show this help message and exit
  -i INPUT, --input INPUT
                        Path to input video to stabilize.
  -o OUTPUT, --output OUTPUT
                        Path to save stabilized video.
  -p PLAYBACK, --playback PLAYBACK
                        Should stabilization be played to screen? (y/n)
  -k KEYPOINTMETHOD, --keyPointMethod KEYPOINTMETHOD
                        Name of keypoint detector to use.
  -s SMOOTHWINDOW, --smoothWindow SMOOTHWINDOW
                        Smoothing window to use while smoothing frame
                        transforms.
  -m MAXFRAMES, --maxFrames MAXFRAMES
                        Max frames to process in video. Negative values will
                        not apply limit.
  -b BORDERTYPE, --borderType BORDERTYPE
                        How to fill in empty border caused by frame shifting.
                        Options: ['black', 'reflect', 'replicate']
  -z BORDERSIZE, --borderSize BORDERSIZE
                        If positive, borderType is added equal to borderSize.
                        If negative, cropping is applied. If 'auto', auto
                        sizing is used to fit transformations.
  -l LAYERFRAMES, --layerFrames LAYERFRAMES
                        Should frame layering effect be applied to output
                        video? (y/n)
"""

if __name__ == '__main__':
    import argparse
    from .main_utils import cli_stabilizer, str_int, str_2_bool

    ap = argparse.ArgumentParser()
    ap.add_argument('-i', '--input', type=str_int, required=True,
                    help='Path to input video to stabilize.')
    ap.add_argument('-o', '--output', required=True,
                    help='Path to save stabilized video.')
    ap.add_argument('-p', '--playback', type=str_2_bool, default='false',
                    help='Should stabilization be played to screen? (y/n)')
    ap.add_argument('-k', '--keyPointMethod', default='GFTT',
                    help='Name of keypoint detector to use.')

    ap.add_argument('-s', '--smoothWindow', default=30, type=int,
                    help='Smoothing window to use while smoothing frame transforms.')
    ap.add_argument('-m', '--maxFrames', default=-1, type=int,
                    help='Max frames to process in video. Negative values will not apply limit.')
    ap.add_argument('-b', '--borderType', default='black',
                    help="How to fill in empty border caused by frame shifting. "
                         "Options: ['black', 'reflect', 'replicate']")
    ap.add_argument('-z', '--borderSize', default=0, type=str_int,
                    help="If positive, borderType is added equal to borderSize. "
                         "If negative, cropping is applied. "
                         "If 'auto', auto sizing is used to fit transformations.")
    ap.add_argument('-l', '--layerFrames', type=str_2_bool, default='false',
                    help='Should frame layering effect be applied to output video? (y/n)')
    args = vars(ap.parse_args())

    cli_stabilizer(args)



import imutils
import cv2


def safe_import_cv2():
    """Gracefully fail ModuleNotFoundError due to cv2

    :return: None
    """
    try:
        import cv2
    except ModuleNotFoundError:
        raise ModuleNotFoundError("""
        No python bindings for OpenCV found when attempting to `import cv2`.
        If you have not installed OpenCV you can install with:

            pip install vidstab[cv2]

        If you'd prefer to install OpenCV from source you can see the docs here:
            https://docs.opencv.org/3.4.1/da/df6/tutorial_py_table_of_contents_setup.html
        """)


def cv2_estimateRigidTransform(from_pts, to_pts, full=False):
    """Estimate transforms in OpenCV 3 or OpenCV 4"""
    if imutils.is_cv4():
        transform = cv2.estimateAffinePartial2D(from_pts, to_pts)[0]
    else:
        transform = cv2.estimateRigidTransform(from_pts, to_pts, full)

    return transform



from urllib.request import urlretrieve


REMOTE_OSTRICH_VID_PATH = 'https://s3.amazonaws.com/python-vidstab/ostrich.mp4'
REMOTE_TRUNCATED_OSTRICH_VID_PATH = 'https://s3.amazonaws.com/python-vidstab/trunc_video.avi'
REMOTE_SKATELINE_VID_PATH = 'https://s3.amazonaws.com/python-vidstab/thrasher.mp4'


def download_ostrich_video(download_to_path):
    """Download example shaky clip of ostrich used in README (mp4)

    Video used with permission the HappyLiving YouTube channel.
    Original video: https://www.youtube.com/watch?v=9pypPqbV_GM

    :param download_to_path: path to save video to
    :return: None

    >>> from vidstab import VidStab, download_ostrich_video
    >>> path = 'ostrich.mp4'
    >>> download_ostrich_video(path)
    >>>
    >>> stabilizer = VidStab()
    >>> stabilizer.stabilize(path, OUTPUT_VIDEO_PATH)
    """
    urlretrieve(REMOTE_OSTRICH_VID_PATH, download_to_path)


def download_skateline_video(download_to_path=None):
    """Download additional testing video

    NOT FOR GENERAL USE; VIDEO MIGHT BE REMOVED WITHOUT WARNING

    :param download_to_path: path to save video to
    :return: None
    """
    urlretrieve(REMOTE_SKATELINE_VID_PATH, download_to_path)


def download_truncated_ostrich_video(download_to_path=None):
    """Download additional testing video

    NOT FOR GENERAL USE; VIDEO MIGHT BE REMOVED WITHOUT WARNING

    :param download_to_path: path to save video to
    :return: None
    """
    urlretrieve(REMOTE_TRUNCATED_OSTRICH_VID_PATH, download_to_path)



import pytest
import numpy as np

from vidstab.frame_queue import FrameQueue


def test_reset_queue():
    frame_queue = FrameQueue(max_len=10, max_frames=20)
    frame_queue.i = 42

    frame_queue.reset_queue(max_len=30, max_frames=40)

    assert frame_queue.max_len == 30
    assert frame_queue.max_frames == 40
    assert len(frame_queue.frames) == 0
    assert len(frame_queue.inds) == 0
    assert frame_queue.i is None


def test_set_frame_source():
    frame_queue = FrameQueue()

    with pytest.raises(TypeError) as err:
        frame_queue.set_frame_source('fake')

    assert 'Not yet support for non cv2.VideoCapture frame source.' in str(err.value)


def test_read_frame():
    frame_queue = FrameQueue()
    black_frame_image = np.zeros((3, 3, 3), dtype='uint8')

    i, break_flag = frame_queue.read_frame(array=black_frame_image)

    assert i == 0
    assert break_flag is None



import pytest
import numpy as np
from vidstab.frame import Frame


GRAY_FRAME_IMAGE = np.zeros((1, 1), dtype='uint8')
BGR_FRAME_IMAGE = np.zeros((1, 1, 3), dtype='uint8')
BGRA_FRAME_IMAGE = np.array([[[0, 0, 0, 255]]], dtype='uint8')


def test_set_color_format():
    frame_gray = Frame(GRAY_FRAME_IMAGE)
    frame_bgr = Frame(BGR_FRAME_IMAGE)
    frame_bgra = Frame(BGRA_FRAME_IMAGE)
    frame_fake = Frame(np.zeros((1, 2)), color_format='fake')

    assert frame_gray.color_format == 'GRAY'
    assert frame_bgr.color_format == 'BGR'
    assert frame_bgra.color_format == 'BGRA'
    assert frame_fake.color_format == 'fake'

    with pytest.raises(ValueError) as err:
        Frame(np.zeros((1, 2, 42)))

    assert 'Unexpected frame image shape: (1, 2, 42)' in str(err.value)


def test_cvt_color():
    frame_gray = Frame(GRAY_FRAME_IMAGE)
    frame_bgr = Frame(BGR_FRAME_IMAGE)
    frame_bgra = Frame(BGRA_FRAME_IMAGE)

    assert np.alltrue(frame_gray.image == frame_gray.gray_image)
    assert np.alltrue(frame_gray.image == frame_bgr.gray_image)
    assert np.alltrue(frame_gray.image == frame_bgra.gray_image)

    assert np.alltrue(frame_bgr.image == frame_gray.bgr_image)
    assert np.alltrue(frame_bgr.image == frame_bgr.bgr_image)
    assert np.alltrue(frame_bgr.image == frame_bgra.bgr_image)

    assert np.alltrue(frame_bgra.image == frame_gray.bgra_image)
    assert np.alltrue(frame_bgra.image == frame_bgr.bgra_image)
    assert np.alltrue(frame_bgra.image == frame_bgra.bgra_image)



import os
import pickle
from urllib.request import urlopen
import imutils


def download_pickled_transforms(window_size=30, cv4=False):
    base_url = 'https://s3.amazonaws.com/python-vidstab'

    window_size = str(window_size) + '_cv4' if cv4 else window_size

    base_urls = [
        f'{base_url}/ostrich_transforms_{window_size}.pickle',
        f'{base_url}/np_ostrich_trajectory_{window_size}.pickle',
        f'{base_url}/np_ostrich_smooth_trajectory_{window_size}.pickle'
    ]

    objs = [pickle_load_from_url(url) for url in base_urls]

    return tuple(objs)


def pickle_load_from_url(url):
    with urlopen(url) as f:
        unpickled_obj = pickle.load(f)

    return unpickled_obj


def pickle_dump(obj, file_path):
    with open(file_path, 'wb') as handle:
        pickle.dump(obj, handle, protocol=pickle.HIGHEST_PROTOCOL)


def pickle_test_transforms(vidstab_obj, path):
    suffix = '_cv4.pickle' if imutils.is_cv4() else '.pickle'

    if not os.path.exists(path):
        os.makedirs(path)

    base_paths = [
        '{}/ostrich_transforms_{}{}',
        '{}/np_ostrich_trajectory_{}{}',
        '{}/np_ostrich_smooth_trajectory_{}{}'
    ]

    paths = [p.format(path, vidstab_obj._smoothing_window, suffix) for p in base_paths]

    pickle_dump(vidstab_obj.transforms, paths[0])
    pickle_dump(vidstab_obj.trajectory, paths[1])
    pickle_dump(vidstab_obj.smoothed_trajectory, paths[2])



import tempfile
import cv2
import imutils
import numpy as np
import pytest

from vidstab import VidStab
from vidstab.download_videos import download_ostrich_video, download_truncated_ostrich_video
from .pickled_transforms import download_pickled_transforms, pickle_test_transforms

# excluding non-free "SIFT" & "SURF" methods do to exclusion from opencv-contrib-python
# see: https://github.com/skvark/opencv-python/issues/126
kp_methods = ["GFTT", "BRISK", "DENSE", "FAST", "HARRIS", "MSER", "ORB", "STAR"]

tmp_dir = tempfile.TemporaryDirectory()

truncated_ostrich_video = '{}/trunc_vid.avi'.format(tmp_dir.name)
ostrich_video = '{}/vid.avi'.format(tmp_dir.name)

download_truncated_ostrich_video(truncated_ostrich_video)
download_ostrich_video(ostrich_video)


# test that all keypoint detection methods load without error
def test_default_init():
    for kp in kp_methods:
        print('testing kp method {}'.format(kp))
        assert VidStab(kp_method=kp).kp_method == kp


def test_kp_options():
    stabilizer = VidStab(kp_method='FAST', threshold=42, nonmaxSuppression=False)
    assert not stabilizer.kp_detector.getNonmaxSuppression()
    assert stabilizer.kp_detector.getThreshold() == 42

    with pytest.raises(TypeError) as err:
        VidStab(kp_method='FAST', fake='fake')

    assert 'invalid keyword argument' in str(err.value)


def test_video_dep_funcs_run():
    # just tests to check functions run
    stabilizer = VidStab()
    stabilizer.gen_transforms(truncated_ostrich_video, smoothing_window=2, show_progress=True)

    assert stabilizer.smoothed_trajectory.shape == stabilizer.trajectory.shape
    assert stabilizer.transforms.shape == stabilizer.trajectory.shape

    with tempfile.TemporaryDirectory() as tmpdir:
        output_vid = '{}/test_output.avi'.format(tmpdir)
        stabilizer.apply_transforms(truncated_ostrich_video, output_vid)
        stabilizer.stabilize(truncated_ostrich_video, output_vid, smoothing_window=2)


def test_trajectory_transform_values():
    for window in [15, 30, 60]:
        stabilizer = VidStab()
        stabilizer.gen_transforms(input_path=ostrich_video, smoothing_window=window)

        pickle_test_transforms(stabilizer, 'pickled_transforms')

        unpickled_transforms = download_pickled_transforms(window, cv4=imutils.is_cv4())

        assert np.allclose(stabilizer.transforms, unpickled_transforms[0])
        assert np.allclose(stabilizer.trajectory, unpickled_transforms[1])
        assert np.allclose(stabilizer.smoothed_trajectory, unpickled_transforms[2])


def test_stabilize_frame():
    # Init stabilizer and video reader
    stabilizer = VidStab()
    vidcap = cv2.VideoCapture(ostrich_video)

    window_size = 30
    while True:
        grabbed_frame, frame = vidcap.read()

        # Pass frame to stabilizer even if frame is None
        stabilized_frame = stabilizer.stabilize_frame(input_frame=frame,
                                                      smoothing_window=window_size,
                                                      border_size=10)

        if stabilized_frame is None:
            break

    unpickled_transforms = download_pickled_transforms(window_size, cv4=imutils.is_cv4())

    assert np.allclose(stabilizer.transforms, unpickled_transforms[0])
    assert np.allclose(stabilizer.trajectory, unpickled_transforms[1])
    assert np.allclose(stabilizer.smoothed_trajectory, unpickled_transforms[2])



from collections import deque


class PopDeque(deque):
    def deque_full(self):
        """Test if queue is full"""
        return len(self) == self.maxlen

    def pop_append(self, x):
        """deque.append helper to return popped element if deque is at ``maxlen``

        :param x: element to append
        :return: result of ``deque.popleft()`` if deque is full; else ``None``

        >>> x = PopDeque([0], maxlen=2)
        >>> x.pop_append(1)

        >>> x.pop_append(2)
        0
        """
        popped_element = None
        if self.deque_full():
            popped_element = self.popleft()

        self.append(x)

        return popped_element

    def increment_append(self, increment=1, pop_append=True):
        """Append deque[-1] + ``increment`` to end of deque

        If deque is empty then 0 is appended

        :param increment: size of increment in deque
        :param pop_append: return popped element if append forces pop?
        :return: popped_element if pop_append is True; else None
        """
        if len(self) == 0:
            popped_element = self.pop_append(0)
        else:
            popped_element = self.pop_append(self[-1] + increment)

        if not pop_append:
            return None

        return popped_element



import matplotlib.pyplot as plt
import numpy as np


def plot_trajectory(transforms, trajectory, smoothed_trajectory):
    """Plot video trajectory

    Create a plot of the video's trajectory & smoothed trajectory.
    Separate subplots are used to show the x and y trajectory.

    :param transforms: VidStab transforms attribute
    :param trajectory: VidStab trajectory attribute
    :param smoothed_trajectory: VidStab smoothed_trajectory attribute
    :return: tuple of matplotlib objects ``(Figure, (AxesSubplot, AxesSubplot))``

    >>> from vidstab import VidStab
    >>> import matplotlib.pyplot as plt
    >>> stabilizer = VidStab()
    >>> stabilizer.gen_transforms(input_path='input_video.mov')
    >>> stabilizer.plot_trajectory()
    >>> plt.show()
    """
    if transforms is None:
        raise AttributeError('No trajectory to plot. '
                             'Use methods: gen_transforms or stabilize to generate the trajectory attributes')

    with plt.style.context('ggplot'):
        fig, (ax1, ax2) = plt.subplots(2, sharex='all')

        # x trajectory
        ax1.plot(trajectory[:, 0], label='Trajectory')
        ax1.plot(smoothed_trajectory[:, 0], label='Smoothed Trajectory')
        ax1.set_ylabel('dx')

        # y trajectory
        ax2.plot(trajectory[:, 1], label='Trajectory')
        ax2.plot(smoothed_trajectory[:, 1], label='Smoothed Trajectory')
        ax2.set_ylabel('dy')

        handles, labels = ax2.get_legend_handles_labels()
        fig.legend(handles, labels, loc='upper right')

        plt.xlabel('Frame Number')

        fig.suptitle('Video Trajectory', x=0.15, y=0.96, ha='left')
        fig.canvas.set_window_title('Trajectory')

        return fig, (ax1, ax2)


def plot_transforms(transforms, radians=False):
    """Plot stabilizing transforms

    Create a plot of the transforms used to stabilize the input video.
    Plots x & y transforms (dx & dy) in a separate subplot than angle transforms (da).

    :param transforms: VidStab transforms attribute
    :param radians: Should angle transforms be plotted in radians?  If ``false``, transforms are plotted in degrees.
    :return: tuple of matplotlib objects ``(Figure, (AxesSubplot, AxesSubplot))``

    >>> from vidstab import VidStab
    >>> import matplotlib.pyplot as plt
    >>> stabilizer = VidStab()
    >>> stabilizer.gen_transforms(input_path='input_video.mov')
    >>> stabilizer.plot_transforms()
    >>> plt.show()
    """
    if transforms is None:
        raise AttributeError('No transforms to plot. '
                             'Use methods: gen_transforms or stabilize to generate the transforms attribute')

    with plt.style.context('ggplot'):
        fig, (ax1, ax2) = plt.subplots(2, sharex='all')

        ax1.plot(transforms[:, 0], label='delta x', color='C0')
        ax1.plot(transforms[:, 1], label='delta y', color='C1')
        ax1.set_ylabel('Delta Pixels', fontsize=10)

        if radians:
            ax2.plot(transforms[:, 2], label='delta angle', color='C2')
            ax2.set_ylabel('Delta Radians', fontsize=10)
        else:
            ax2.plot(np.rad2deg(transforms[:, 2]), label='delta angle', color='C2')
            ax2.set_ylabel('Delta Degrees', fontsize=10)

        handles1, labels1 = ax1.get_legend_handles_labels()
        handles2, labels2 = ax2.get_legend_handles_labels()
        fig.legend(handles1 + handles2,
                   labels1 + labels2,
                   loc='upper right',
                   ncol=1)

        plt.xlabel('Frame Number')

        fig.suptitle('Transformations for Stabilizing', x=0.15, y=0.96, ha='left')
        fig.canvas.set_window_title('Transforms')

        return fig, (ax1, ax2)



"""VidStab: a class for stabilizing video files"""

from .cv2_utils import safe_import_cv2
safe_import_cv2()  # inform user of pip install vidstab[cv2] if ModuleNotFoundError

import time
import warnings
import cv2
import numpy as np
import imutils.feature.factories as kp_factory
import matplotlib.pyplot as plt
from . import general_utils
from . import vidstab_utils
from . import border_utils
from . import auto_border_utils
from . import plot_utils
from .frame_queue import FrameQueue
from . frame import Frame


class VidStab:
    """A class for stabilizing video files

    The VidStab class can be used to stabilize videos using functionality from OpenCV.
    Input video is read from file, put through stabilization process, and written to
    an output file.

    The process calculates optical flow (``cv2.calcOpticalFlowPyrLK``) from frame to frame using
    keypoints generated by the keypoint method specified by the user.  The optical flow will
    be used to generate frame to frame transformations (``cv2.estimateRigidTransform``).
    Transformations will be applied (``cv2.warpAffine``) to stabilize video.

    This class is based on the `work presented by Nghia Ho <http://nghiaho.com/?p=2093>`_

    :param kp_method: String of the type of keypoint detector to use. Available options are:
                        ``["GFTT", "BRISK", "DENSE", "FAST", "HARRIS", "MSER", "ORB", "STAR"]``.
                        ``["SIFT", "SURF"]`` are additional non-free options available depending
                        on your build of OpenCV.  The non-free detectors are not tested with this package.
    :param args: Positional arguments for keypoint detector.
    :param kwargs: Keyword arguments for keypoint detector.

    :ivar kp_method: a string naming the keypoint detector being used
    :ivar kp_detector: the keypoint detector object being used
    :ivar trajectory: a 2d showing the trajectory of the input video
    :ivar smoothed_trajectory: a 2d numpy array showing the smoothed trajectory of the input video
    :ivar transforms: a 2d numpy array storing the transformations used from frame to frame
    """

    def __init__(self, kp_method='GFTT', *args, **kwargs):
        """instantiate VidStab class

        :param kp_method: String of the type of keypoint detector to use. Available options are:
                        ``["GFTT", "BRISK", "DENSE", "FAST", "HARRIS", "MSER", "ORB", "STAR"]``.
                        ``["SIFT", "SURF"]`` are additional non-free options available depending
                        on your build of OpenCV.  The non-free detectors are not tested with this package.
        :param args: Positional arguments for keypoint detector.
        :param kwargs: Keyword arguments for keypoint detector.
        """

        self.kp_method = kp_method
        # use original defaults in http://nghiaho.com/?p=2093 if GFTT with no additional (kw)args
        if kp_method == 'GFTT' and args == () and kwargs == {}:
            self.kp_detector = kp_factory.FeatureDetector_create('GFTT',
                                                                 maxCorners=200,
                                                                 qualityLevel=0.01,
                                                                 minDistance=30.0,
                                                                 blockSize=3)
        else:
            self.kp_detector = kp_factory.FeatureDetector_create(kp_method, *args, **kwargs)

        self._smoothing_window = 30
        self._raw_transforms = []
        self._trajectory = []
        self.trajectory = self.smoothed_trajectory = self.transforms = None

        self.frame_queue = FrameQueue()
        self.prev_kps = self.prev_gray = None

        self.writer = None

        self.layer_options = {
            'layer_func': None,
            'prev_frame': None
        }

        self.border_options = {}
        self.auto_border_flag = False
        self.extreme_frame_corners = {'min_x': 0, 'min_y': 0, 'max_x': 0, 'max_y': 0}
        self.frame_corners = None

        self._default_stabilize_frame_output = None

    def _update_prev_frame(self, current_frame_gray):
        self.prev_gray = current_frame_gray[:]
        self.prev_kps = self.kp_detector.detect(self.prev_gray)
        # noinspection PyArgumentList
        self.prev_kps = np.array([kp.pt for kp in self.prev_kps], dtype='float32').reshape(-1, 1, 2)

    def _update_trajectory(self, transform):
        if not self._trajectory:
            self._trajectory.append(transform[:])
        else:
            # gen cumsum for new row and append
            self._trajectory.append([self._trajectory[-1][j] + x for j, x in enumerate(transform)])

    def _gen_next_raw_transform(self):
        current_frame = self.frame_queue.frames[-1]
        current_frame_gray = current_frame.gray_image

        # calc flow of movement
        optical_flow = cv2.calcOpticalFlowPyrLK(self.prev_gray,
                                                current_frame_gray,
                                                self.prev_kps, None)

        matched_keypoints = vidstab_utils.match_keypoints(optical_flow, self.prev_kps)
        transform_i = vidstab_utils.estimate_partial_transform(matched_keypoints)

        # update previous frame info for next iteration
        self._update_prev_frame(current_frame_gray)
        self._raw_transforms.append(transform_i[:])
        self._update_trajectory(transform_i)

    def _init_is_complete(self, gen_all):
        if gen_all:
            return False

        max_ind = min([self.frame_queue.max_frames,
                       self.frame_queue.max_len])

        if self.frame_queue.inds[-1] >= max_ind:
            return True

        return False

    def _process_first_frame(self, array=None):
        # read first frame
        _, _ = self.frame_queue.read_frame(array=array)
        # convert to gray scale
        prev_frame = self.frame_queue.frames[-1]
        prev_frame_gray = prev_frame.gray_image
        # detect keypoints
        prev_kps = self.kp_detector.detect(prev_frame_gray)
        # noinspection PyArgumentList
        self.prev_kps = np.array([kp.pt for kp in prev_kps], dtype='float32').reshape(-1, 1, 2)

        self.prev_gray = prev_frame_gray[:]

    def _init_trajectory(self, smoothing_window, max_frames, gen_all=False, show_progress=False):
        self._smoothing_window = smoothing_window

        if max_frames is None:
            max_frames = float('inf')

        frame_count = self.frame_queue.source_frame_count
        bar = general_utils.init_progress_bar(frame_count, max_frames, show_progress, gen_all)

        self._process_first_frame()

        # iterate through frames
        while True:
            # read current frame
            _, break_flag = self.frame_queue.read_frame(pop_ind=False)
            if not self.frame_queue.grabbed_frame:
                general_utils.update_progress_bar(bar, show_progress)
                break

            self._gen_next_raw_transform()

            if self._init_is_complete(gen_all):
                break

            general_utils.update_progress_bar(bar, show_progress)

        self._gen_transforms()

        return bar

    def _init_writer(self, output_path, frame_shape, output_fourcc, fps):
        # set output and working dims
        h, w = frame_shape

        # setup video writer
        self.writer = cv2.VideoWriter(output_path,
                                      cv2.VideoWriter_fourcc(*output_fourcc),
                                      fps, (w, h), True)

    def _set_border_options(self, border_size, border_type):
        functional_border_size, functional_neg_border_size = border_utils.functional_border_sizes(border_size)

        self.border_options = {
            'border_type': border_type,
            'border_size': functional_border_size,
            'neg_border_size': functional_neg_border_size,
            'extreme_frame_corners': self.extreme_frame_corners,
            'auto_border_flag': self.auto_border_flag
        }

    def _apply_transforms(self, output_path, max_frames, use_stored_transforms,
                          output_fourcc='MJPG', border_type='black', border_size=0,
                          layer_func=None, playback=False, progress_bar=None):

        self._set_border_options(border_size, border_type)
        self.layer_options['layer_func'] = layer_func

        while True:
            general_utils.update_progress_bar(progress_bar)
            i, break_flag = self.frame_queue.read_frame()

            if not self.frame_queue.frames_to_process() or break_flag:
                break

            if not use_stored_transforms:
                self._gen_next_raw_transform()

            transformed = self._apply_next_transform(i, use_stored_transforms=use_stored_transforms)

            if transformed is None:
                warnings.warn('Video is longer than available transformations; halting process.')
                break

            break_playback = general_utils.playback_video(transformed, playback,
                                                          delay=min([self._smoothing_window, max_frames]))
            if break_playback:
                break

            if self.writer is None:
                self._init_writer(output_path, transformed.shape[:2], output_fourcc,
                                  fps=self.frame_queue.source_fps)

            self.writer.write(transformed)

        self.writer.release()
        general_utils.update_progress_bar(progress_bar, finish=True)
        cv2.destroyAllWindows()

    def _gen_transforms(self):
        self.trajectory = np.array(self._trajectory)
        self.smoothed_trajectory = general_utils.bfill_rolling_mean(self.trajectory, n=self._smoothing_window)
        self.transforms = np.array(self._raw_transforms) + (self.smoothed_trajectory - self.trajectory)

    def gen_transforms(self, input_path, smoothing_window=30, show_progress=True):
        """Generate stabilizing transforms for a video

        This method will populate the following instance attributes: trajectory, smoothed_trajectory, & transforms.
        The resulting transforms can subsequently be used for video stabilization by using ``VidStab.apply_transforms``
        or ``VidStab.stabilize`` with ``use_stored_transforms=True``.

        :param input_path: Path to input video to stabilize.
                           Will be read with ``cv2.VideoCapture``; see opencv documentation for more info.
        :param smoothing_window: window size to use when smoothing trajectory
        :param show_progress: Should a progress bar be displayed to console?
        :return: Nothing; this method populates attributes of VidStab objects

        >>> from vidstab.VidStab import VidStab
        >>> stabilizer = VidStab()
        >>> stabilizer.gen_transforms(input_path='input_video.mov')
        >>> stabilizer.apply_transforms(input_path='input_video.mov', output_path='stable_video.avi')
        """
        self._smoothing_window = smoothing_window
        self.frame_queue.set_frame_source(cv2.VideoCapture(input_path))
        self.frame_queue.reset_queue(max_len=smoothing_window, max_frames=float('inf'))
        bar = self._init_trajectory(smoothing_window=smoothing_window,
                                    max_frames=float('inf'),
                                    gen_all=True,
                                    show_progress=show_progress)

        general_utils.update_progress_bar(bar, finish=True)

    def apply_transforms(self, input_path, output_path, output_fourcc='MJPG',
                         border_type='black', border_size=0, layer_func=None, show_progress=True, playback=False):
        """Apply stored transforms to a video and save output to file

        Use the transforms generated by ``VidStab.gen_transforms`` or ``VidStab.stabilize`` in stabilization process.
        This method is a wrapper for ``VidStab.stabilize`` with ``use_stored_transforms=True``;
        it is included for backwards compatibility.

        :param input_path: Path to input video to stabilize.
                           Will be read with ``cv2.VideoCapture``; see opencv documentation for more info.
        :param output_path: Path to save stabilized video.
                            Will be written with ``cv2.VideoWriter``; see opencv documentation for more info.
        :param output_fourcc: FourCC is a 4-byte code used to specify the video codec.
        :param border_type: How to handle negative space created by stabilization translations/rotations.
                            Options: ``['black', 'reflect', 'replicate']``
        :param border_size: Size of border in output.
                            Positive values will pad video equally on all sides,
                            negative values will crop video equally on all sides,
                            ``'auto'`` will attempt to minimally pad to avoid cutting off portions of transformed frames
        :param layer_func: Function to layer frames in output.
                           The function should accept 2 parameters: foreground & background.
                           The current frame of video will be passed as foreground,
                           the previous frame will be passed as the background
                           (after the first frame of output the background will be the output of
                           layer_func on the last iteration)
        :param show_progress: Should a progress bar be displayed to console?
        :param playback: Should the a comparison of input video/output video be played back during process?
        :return: Nothing is returned.  Output of stabilization is written to ``output_path``.

        >>> from vidstab.VidStab import VidStab
        >>> stabilizer = VidStab()
        >>> stabilizer.gen_transforms(input_path='input_video.mov')
        >>> stabilizer.apply_transforms(input_path='input_video.mov', output_path='stable_video.avi')
        """
        self.stabilize(input_path, output_path, smoothing_window=self._smoothing_window, max_frames=float('inf'),
                       border_type=border_type, border_size=border_size, layer_func=layer_func, playback=playback,
                       use_stored_transforms=True, show_progress=show_progress, output_fourcc=output_fourcc)

    def _apply_next_transform(self, i, use_stored_transforms=False):
        if not use_stored_transforms:
            self._gen_transforms()

        frame_i = self.frame_queue.frames.popleft()

        try:
            transform_i = self.transforms[i, :]
        except IndexError:
            return None

        transformed = vidstab_utils.transform_frame(frame_i,
                                                    transform_i,
                                                    self.border_options['border_size'],
                                                    self.border_options['border_type'])

        transformed, self.layer_options = vidstab_utils.post_process_transformed_frame(transformed,
                                                                                       self.border_options,
                                                                                       self.layer_options)

        transformed = transformed.cvt_color(frame_i.color_format)

        return transformed

    def stabilize_frame(self, input_frame, smoothing_window=30,
                        border_type='black', border_size=0, layer_func=None,
                        use_stored_transforms=False):
        """Stabilize single frame of video being iterated

        Perform video stabilization a single frame at a time.  Outputted stabilized frame will be on a
        ``smoothing_window`` delay.  When frames processed is ``< smoothing_window``, black frames will be returned.
        When frames processed is ``>= smoothing_window``, the stabilized frame ``smoothing_window`` ago will be
        returned.  When ``input_frame is None`` stabilization will still be attempted, if there are not frames left to
        process then ``None`` will be returned.

        :param input_frame: An OpenCV image (as numpy array) or None
        :param smoothing_window: window size to use when smoothing trajectory
        :param border_type: How to handle negative space created by stabilization translations/rotations.
                            Options: ``['black', 'reflect', 'replicate']``
        :param border_size: Size of border in output.
                            Positive values will pad video equally on all sides,
                            negative values will crop video equally on all sides,
                            ``'auto'`` will attempt to minimally pad to avoid cutting off portions of transformed frames
        :param layer_func: Function to layer frames in output.
                           The function should accept 2 parameters: foreground & background.
                           The current frame of video will be passed as foreground,
                           the previous frame will be passed as the background
                           (after the first frame of output the background will be the output of
                           layer_func on the last iteration)
        :param use_stored_transforms: should stored transforms from last stabilization be used instead of
                                      recalculating them?
        :return: 1 of 3 outputs will be returned:

            * Case 1 - Stabilization process is still warming up
                + **An all black frame of same shape as input_frame is returned.**
                + A minimum of ``smoothing_window`` frames need to be processed to perform stabilization.
                + This behavior was based on ``cv2.bgsegm.createBackgroundSubtractorMOG()``.
            * Case 2 - Stabilization process is warmed up and ``input_frame is not None``
                + **A stabilized frame is returned**
                + This will not be the stabilized version of ``input_frame``.
                  Stabilization is on an ``smoothing_window`` frame delay
            * Case 3 - Stabilization process is finished
                + **None**

        >>> from vidstab.VidStab import VidStab
        >>> stabilizer = VidStab()
        >>> vidcap = cv2.VideoCapture('input_video.mov')
        >>> while True:
        >>>     grabbed_frame, frame = vidcap.read()
        >>>     # Pass frame to stabilizer even if frame is None
        >>>     # stabilized_frame will be an all black frame until iteration 30
        >>>     stabilized_frame = stabilizer.stabilize_frame(input_frame=frame,
        >>>                                                   smoothing_window=30)
        >>>     if stabilized_frame is None:
        >>>         # There are no more frames available to stabilize
        >>>         break
        """
        self._set_border_options(border_size, border_type)
        self.layer_options['layer_func'] = layer_func

        # Store first frame
        if self.frame_queue.max_len is None:
            self.frame_queue.reset_queue(max_len=smoothing_window, max_frames=float('inf'))

            self._process_first_frame(array=input_frame)

            blank_frame = Frame(np.zeros_like(input_frame))
            blank_frame = border_utils.crop_frame(blank_frame, self.border_options)

            if self.border_options['border_size'] > 0:
                blank_frame_alpha, _ = vidstab_utils.border_frame(blank_frame,
                                                                  self.border_options['border_size'],
                                                                  self.border_options['border_type'])

                blank_frame_np = Frame(blank_frame_alpha).cvt_color(blank_frame.color_format)
                blank_frame = Frame(blank_frame_np)

            self._default_stabilize_frame_output = blank_frame.image

            return self._default_stabilize_frame_output

        if len(self.frame_queue.frames) == 0:
            return None

        if input_frame is not None:
            _, _ = self.frame_queue.read_frame(array=input_frame, pop_ind=False)
            if not use_stored_transforms:
                self._gen_next_raw_transform()

        if not self._init_is_complete(gen_all=False):
            return self._default_stabilize_frame_output

        stabilized_frame = self._apply_next_transform(self.frame_queue.i, use_stored_transforms=use_stored_transforms)

        return stabilized_frame

    def stabilize(self, input_path, output_path, smoothing_window=30, max_frames=float('inf'),
                  border_type='black', border_size=0, layer_func=None, playback=False,
                  use_stored_transforms=False, show_progress=True, output_fourcc='MJPG'):
        """Read video, perform stabilization, & write stabilized video to file

        :param input_path: Path to input video to stabilize.
                           Will be read with ``cv2.VideoCapture``; see opencv documentation for more info.
        :param output_path: Path to save stabilized video.
                            Will be written with ``cv2.VideoWriter``; see opencv documentation for more info.
        :param smoothing_window: window size to use when smoothing trajectory
        :param max_frames: The maximum amount of frames to stabilize/process.
                           The list of available codes can be found in fourcc.org.
                           See cv2.VideoWriter_fourcc documentation for more info.
        :param border_type: How to handle negative space created by stabilization translations/rotations.
                            Options: ``['black', 'reflect', 'replicate']``
        :param border_size: Size of border in output.
                            Positive values will pad video equally on all sides,
                            negative values will crop video equally on all sides,
                            ``'auto'`` will attempt to minimally pad to avoid cutting off portions of transformed frames
        :param layer_func: Function to layer frames in output.
                           The function should accept 2 parameters: foreground & background.
                           The current frame of video will be passed as foreground,
                           the previous frame will be passed as the background
                           (after the first frame of output the background will be the output of
                           layer_func on the last iteration)
        :param use_stored_transforms: should stored transforms from last stabilization be used instead of
                                      recalculating them?
        :param playback: Should the a comparison of input video/output video be played back during process?
        :param show_progress: Should a progress bar be displayed to console?
        :param output_fourcc: FourCC is a 4-byte code used to specify the video codec.
        :return: Nothing is returned.  Output of stabilization is written to ``output_path``.

        >>> from vidstab.VidStab import VidStab
        >>> stabilizer = VidStab()
        >>> stabilizer.stabilize(input_path='input_video.mov', output_path='stable_video.avi')

        >>> stabilizer = VidStab(kp_method = 'ORB')
        >>> stabilizer.stabilize(input_path='input_video.mov', output_path='stable_video.avi')
        """
        if border_size == 'auto':
            self.auto_border_flag = True

        self.frame_queue.set_frame_source(cv2.VideoCapture(input_path))

        # wait for camera to start up
        if isinstance(input_path, int):
            time.sleep(0.1)

        self.frame_queue.reset_queue(max_len=smoothing_window, max_frames=max_frames)

        if self.auto_border_flag and not use_stored_transforms:
            use_stored_transforms = True
            self.gen_transforms(input_path, smoothing_window=smoothing_window, show_progress=show_progress)
            self.frame_queue.set_frame_source(cv2.VideoCapture(input_path))
            self.frame_queue.reset_queue(max_len=smoothing_window, max_frames=max_frames)

            bar = general_utils.init_progress_bar(self.frame_queue.source_frame_count,
                                                  max_frames,
                                                  show_progress)
            self.frame_queue.populate_queue(smoothing_window)

        elif not use_stored_transforms:
            bar = self._init_trajectory(smoothing_window, max_frames, show_progress=show_progress)
        else:
            bar = general_utils.init_progress_bar(self.frame_queue.source_frame_count,
                                                  max_frames,
                                                  show_progress)
            self.frame_queue.populate_queue(smoothing_window)

        if self.auto_border_flag:
            frame_1 = self.frame_queue.frames[0]
            self.extreme_frame_corners = auto_border_utils.extreme_corners(frame_1.image, self.transforms)
            border_size = auto_border_utils.min_auto_border_size(self.extreme_frame_corners)

        self._apply_transforms(output_path, max_frames, use_stored_transforms=use_stored_transforms,
                               border_type=border_type, border_size=border_size,
                               layer_func=layer_func, playback=playback,
                               output_fourcc=output_fourcc, progress_bar=bar)

    def plot_trajectory(self):
        """Plot video trajectory

        Create a plot of the video's trajectory & smoothed trajectory.
        Separate subplots are used to show the x and y trajectory.

        :return: tuple of matplotlib objects ``(Figure, (AxesSubplot, AxesSubplot))``

        >>> from vidstab import VidStab
        >>> import matplotlib.pyplot as plt
        >>> stabilizer = VidStab()
        >>> stabilizer.gen_transforms(input_path='input_video.mov')
        >>> stabilizer.plot_trajectory()
        >>> plt.show()
        """
        return plot_utils.plot_trajectory(self.transforms, self.trajectory, self.smoothed_trajectory)

    def plot_transforms(self, radians=False):
        """Plot stabilizing transforms

        Create a plot of the transforms used to stabilize the input video.
        Plots x & y transforms (dx & dy) in a separate subplot than angle transforms (da).

        :param radians: Should angle transforms be plotted in radians?  If ``false``, transforms are plotted in degrees.
        :return: tuple of matplotlib objects ``(Figure, (AxesSubplot, AxesSubplot))``

        >>> from vidstab import VidStab
        >>> import matplotlib.pyplot as plt
        >>> stabilizer = VidStab()
        >>> stabilizer.gen_transforms(input_path='input_video.mov')
        >>> stabilizer.plot_transforms()
        >>> plt.show()
        """
        return plot_utils.plot_transforms(self.transforms, radians)



from .auto_border_utils import auto_border_crop
from .frame import Frame


def functional_border_sizes(border_size):
    """Calculate border sizing used in process to gen user specified border size

    If border_size is negative then a stand-in border size is used to allow better keypoint tracking (i think... ?);
    negative border is then applied at end.

    :param border_size: user supplied border size
    :return: (border_size, neg_border_size) tuple of functional border sizes

    >>> functional_border_sizes(100)
    (100, 0)
    >>> functional_border_sizes(-10)
    (100, 110)
    """
    if border_size < 0:
        neg_border_size = 100 + abs(border_size)
        border_size = 100
    else:
        neg_border_size = 0

    return border_size, neg_border_size


def crop_frame(frame, border_options):
    """Handle frame cropping for auto border size and negative border size

    if auto_border is False and neg_border_size == 0 then frame is returned as is

    :param frame: frame to be cropped
    :param border_options: dictionary of border options including keys for:
        * 'border_size': functional border size determined by functional_border_sizes
        * 'neg_border_size': functional negative border size determined by functional_border_sizes
        * 'extreme_frame_corners': VidStab.extreme_frame_corners attribute
        * 'auto_border': VidStab.auto_border_flag attribute
    :return: cropped frame
    """
    if not border_options['auto_border_flag'] and border_options['neg_border_size'] == 0:
        return frame

    if border_options['auto_border_flag']:
        cropped_frame_image = auto_border_crop(frame.image,
                                               border_options['extreme_frame_corners'],
                                               border_options['border_size'])

    else:
        frame_h, frame_w = frame.image.shape[:2]
        cropped_frame_image = frame.image[
                              border_options['neg_border_size']:(frame_h - border_options['neg_border_size']),
                              border_options['neg_border_size']:(frame_w - border_options['neg_border_size'])
                              ]

    cropped_frame = Frame(cropped_frame_image, color_format=frame.color_format)

    return cropped_frame



import cv2
from .frame import Frame
from .pop_deque import PopDeque


class FrameQueue:
    def __init__(self, max_len=None, max_frames=None):
        self.max_len = max_len
        self.max_frames = max_frames

        self.frames = PopDeque(maxlen=max_len)
        self.inds = PopDeque(maxlen=max_len)
        self.i = None

        self.source = None
        self.source_frame_count = None
        self.source_fps = 30

        self.grabbed_frame = False

    def reset_queue(self, max_len=None, max_frames=None):
        self.max_len = max_len if max_len is not None else self.max_len
        self.max_frames = max_frames if max_frames is not None else self.max_frames

        self.frames = PopDeque(maxlen=max_len)
        self.inds = PopDeque(maxlen=max_len)
        self.i = None

    def set_frame_source(self, source):
        if isinstance(source, cv2.VideoCapture):
            self.source = source
            self.source_frame_count = int(source.get(cv2.CAP_PROP_FRAME_COUNT))
            self.source_fps = int(source.get(cv2.CAP_PROP_FPS))
        else:
            raise TypeError('Not yet support for non cv2.VideoCapture frame source.')

    def read_frame(self, pop_ind=True, array=None):
        if isinstance(self.source, cv2.VideoCapture):
            self.grabbed_frame, frame = self.source.read()
        else:
            frame = array

        return self._append_frame(frame, pop_ind)

    def _append_frame(self, frame, pop_ind=True):
        if frame is not None:
            self.frames.append(Frame(frame))
            self.i = self.inds.increment_append()

        if pop_ind and self.i is None:
            self.i = self.inds.popleft()

        if (pop_ind
                and self.i is not None
                and self.max_frames is not None):
            break_flag = self.i >= self.max_frames
        else:
            break_flag = None

        return self.i, break_flag

    def populate_queue(self, smoothing_window):
        n = min([smoothing_window, self.max_frames])

        for i in range(n):
            _, _ = self.read_frame(pop_ind=False)
            if not self.grabbed_frame:
                break

    def frames_to_process(self):
        return len(self.frames) > 0 or self.grabbed_frame



import builtins
import sys
import pytest
from vidstab.cv2_utils import safe_import_cv2


@pytest.fixture
def missing_cv2(monkeypatch):
    """Monkey patch import to test missing cv2"""
    import_og = builtins.__import__

    def mocked_import(name, globals, locals, fromlist, level):
        if name == 'cv2':
            raise ModuleNotFoundError()
        return import_og(name, locals, fromlist, level)

    monkeypatch.setattr(builtins, '__import__', mocked_import)


@pytest.fixture(autouse=True)
def cleanup_imports():
    yield
    sys.modules.pop('cv2', None)


@pytest.mark.usefixtures('missing_cv2')
def test_missing_cv2():
    with pytest.raises(ModuleNotFoundError) as err:
        safe_import_cv2()

    assert 'pip install vidstab[cv2]' in str(err.value)



from .VidStab import VidStab
from .download_videos import download_ostrich_video
from .layer_utils import layer_blend, layer_overlay
from .version import __version__

__author__ = "Adam Spannbauer <spannbaueradam@gmail.com>"



import cv2
import numpy as np
from .frame import Frame


def layer_overlay(foreground, background):
    """put an image over the top of another

    Intended for use in VidStab class to create a trail of previous frames in the stable video output.

    :param foreground: image to be laid over top of background image
    :param background: image to over laid with foreground image
    :return: return combined image where foreground is laid over background

    >>> from vidstab import VidStab, layer_overlay, layer_blend
    >>>
    >>> stabilizer = VidStab()
    >>>
    >>> stabilizer.stabilize(input_path=INPUT_VIDEO_PATH,
    >>>                      output_path=OUTPUT_VIDEO_PATH,
    >>>                      border_size=100,
    >>>                      layer_func=layer_overlay)
    """
    overlaid = foreground.copy()
    negative_space = np.where(foreground[:, :, 3] == 0)

    overlaid[negative_space] = background[negative_space]

    overlaid[:, :, 3] = 255

    return overlaid


def layer_blend(foreground, background, foreground_alpha=.6):
    """blend a foreground image over background (wrapper for cv2.addWeighted)

    :param foreground: image to be laid over top of background image
    :param background: image to over laid with foreground image
    :param foreground_alpha: alpha to apply to foreground; (1 - foreground_alpha) applied to background
    :return: return combined image where foreground is laid over background with alpha

    >>> from vidstab import VidStab, layer_overlay, layer_blend
    >>>
    >>> stabilizer = VidStab()
    >>>
    >>> stabilizer.stabilize(input_path=INPUT_VIDEO_PATH,
    >>>                      output_path=OUTPUT_VIDEO_PATH,
    >>>                      border_size=100,
    >>>                      layer_func=layer_blend)
    """
    cv2.addWeighted(foreground, foreground_alpha,
                    background, 1 - foreground_alpha, 0, background)

    return background


def apply_layer_func(cur_frame, prev_frame, layer_func):
    """helper method to apply layering function in vidstab process

    :param cur_frame: current frame to apply layer over prev_frame
    :param prev_frame: previous frame to be layered over by cur_frame
    :param layer_func: layering function to apply
    :return: tuple of (layered_frames, prev_frame) where prev_frame is to be used in next layering operation
    """
    if prev_frame is not None:
        cur_frame_image = layer_func(cur_frame.image, prev_frame.image)
        cur_frame = Frame(cur_frame_image, color_format=cur_frame.color_format)

    return cur_frame



"""
`@image_comparison` was removed in matplotlib 3.2
Tests now a bit looser
"""
import matplotlib
import pytest
from vidstab.plot_utils import plot_transforms, plot_trajectory
from .pickled_transforms import download_pickled_transforms

transforms, trajectory, smooth_trajectory = download_pickled_transforms(window_size=30)


def int_axes_lims(axes):
    x_lims = axes.get_xlim()
    y_lims = axes.get_ylim()

    return [int(x) for x in x_lims + y_lims]


def check_vidstab_plot(plot, expected_axes1, expected_axes2):
    fig, (axes1, axes2) = plot

    axes1_lims = int_axes_lims(axes1)
    axes2_lims = int_axes_lims(axes2)

    fig_check = isinstance(fig, matplotlib.figure.Figure)
    axes1_check = axes1_lims == expected_axes1
    axes2_check = axes2_lims == expected_axes2

    return all([fig_check, axes1_check, axes2_check])


def test_plot_trajectory_exception():
    with pytest.raises(AttributeError) as err:
        plot_trajectory(None, None, None)

    assert 'No trajectory to plot' in str(err.value)


def test_plot_transforms_exception():
    with pytest.raises(AttributeError) as err:
        plot_transforms(None)

    assert 'No transforms to plot' in str(err.value)


def test_plot_trajectory():
    plot = plot_trajectory(transforms, trajectory, smooth_trajectory)

    expected_axes1 = [-15, 325, -384, 173]
    expected_axes2 = [-15, 325, -17, 450]

    assert check_vidstab_plot(plot, expected_axes1, expected_axes2)


def test_plot_transforms():
    plot = plot_transforms(transforms, radians=False)

    expected_axes1 = [-15, 325, -204, 113]
    expected_axes2 = [-15, 325, -10, 5]

    assert check_vidstab_plot(plot, expected_axes1, expected_axes2)


def test_plot_transforms_radians():
    plot = plot_transforms(transforms, radians=True)

    expected_axes1 = [-15, 325, -204, 113]
    expected_axes2 = [-15, 325, 0, 0]

    assert check_vidstab_plot(plot, expected_axes1, expected_axes2)



import numpy as np
import cv2
import vidstab.auto_border_utils as utils
import vidstab.vidstab_utils as vidstab_utils


WHITE_FRAME = 255 * np.ones((100, 200, 3), dtype='uint8')
TEST_TRANSFORMS = np.array([[-69.67148996, 67.21674617, -0.16193986]])
TEST_EXTREME_CORNERS = utils.extreme_corners(WHITE_FRAME, TEST_TRANSFORMS)


def demo_auto_border_crop():
    min_border = utils.min_auto_border_size(TEST_EXTREME_CORNERS)
    h, w = WHITE_FRAME.shape[:2]

    og_frame_corners = np.array([[min_border, min_border],
                                 [min_border + w, min_border],
                                 [min_border, min_border + h],
                                 [min_border + w, min_border + h]])

    transformed_frame_corners = [(min_border - 69, min_border + 67),
                                 (min_border - 53, min_border + 164),
                                 (min_border + 126, min_border + 35),
                                 (min_border + 142, min_border + 132)]

    transformed_frame = vidstab_utils.transform_frame(WHITE_FRAME, TEST_TRANSFORMS[0], min_border, 'black')

    for xy in og_frame_corners:
        cv2.circle(transformed_frame, tuple(xy), 3, (0, 0, 255), -1)

    for xy in transformed_frame_corners:
        cv2.circle(transformed_frame, tuple(xy), 3, (255, 0, 255), -1)

    cropped_transformed_frame = utils.auto_border_crop(transformed_frame, TEST_EXTREME_CORNERS, min_border)

    cv2.imshow('pre-crop (red dots = original frame corners)', transformed_frame)
    cv2.imshow('cropped (red dots = original corners)', cropped_transformed_frame)
    cv2.waitKey(0)


def test_extreme_corners():
    extreme_corners = {k: int(v) for k, v in TEST_EXTREME_CORNERS.items()}

    expected_corners = {
        'min_x': -72,
        'min_y': 0,
        'max_x': 0,
        'max_y': 67
    }

    assert extreme_corners == expected_corners


def test_auto_border_start():
    start = utils.auto_border_start(40.4, 100.6)
    assert isinstance(start, int)
    assert start == 60


def test_auto_border_length():
    length = utils.auto_border_length(400, 42.0, 100.6)
    assert isinstance(length, int)
    assert length == 342


def test_auto_border_crop():
    extreme_frame_corners = {
        'min_x': 0,
        'min_y': 0,
        'max_x': 0,
        'max_y': 0
    }

    cropped_frame = utils.auto_border_crop(WHITE_FRAME, extreme_frame_corners, 0)
    assert cropped_frame.shape == WHITE_FRAME.shape

    extreme_frame_corners = {
        'min_x': -10,
        'min_y': -20,
        'max_x': 5,
        'max_y': 10
    }
    min_border_size = utils.min_auto_border_size(extreme_frame_corners)
    cropped_frame = utils.auto_border_crop(WHITE_FRAME, extreme_frame_corners, min_border_size)
    assert cropped_frame.shape == (90, 175, 3)


def test_min_auto_border_size():
    extreme_frame_corners = {
        'min_x': 0,
        'min_y': 0,
        'max_x': 0,
        'max_y': 0
    }

    min_size = utils.min_auto_border_size(extreme_frame_corners)
    assert min_size == 0

    extreme_frame_corners['min_y'] = 10
    min_size = utils.min_auto_border_size(extreme_frame_corners)
    assert min_size == 10

    extreme_frame_corners['min_x'] = -20
    min_size = utils.min_auto_border_size(extreme_frame_corners)
    assert min_size == 20


if __name__ == '__main__':
    demo_auto_border_crop()



import cv2


class Frame:
    """Utility for easier color format conversions.

    :param image: OpenCV image as numpy array.
    :param color_format: Name of input color format or None.
         If str, the input must use the format that is used in OpenCV's cvtColor code parameter.
         For example, if an image is bgr then input 'BGR' as seen in the cvtColor codes:
        [cv2.COLOR_BGR2GRAY, COLOR_Luv2BGR].
        If None, the color format will be assumed from shape of the image.
        The only possible outcomes of this assumption are: ['GRAY', 'BGR', 'BGRA'].

    :ivar image: input image with possible color format conversions applied
    :ivar color_format: str containing the current color format of image attribute.
    """
    def __init__(self, image, color_format=None):
        self.image = image

        if color_format is None:
            self.color_format = self._guess_color_format()
        else:
            self.color_format = color_format

    def _guess_color_format(self):
        if len(self.image.shape) == 2:
            return 'GRAY'

        elif self.image.shape[2] == 3:
            return 'BGR'

        elif self.image.shape[2] == 4:
            return 'BGRA'

        else:
            raise ValueError(f'Unexpected frame image shape: {self.image.shape}')

    @staticmethod
    def _lookup_color_conversion(from_format, to_format):
        return getattr(cv2, f'COLOR_{from_format}2{to_format}')

    def cvt_color(self, to_format):
        if not self.color_format == to_format:
            color_conversion = self._lookup_color_conversion(from_format=self.color_format,
                                                             to_format=to_format)

            return cv2.cvtColor(self.image, color_conversion)
        else:
            return self.image

    @property
    def gray_image(self):
        return self.cvt_color('GRAY')

    @property
    def bgr_image(self):
        return self.cvt_color('BGR')

    @property
    def bgra_image(self):
        return self.cvt_color('BGRA')



import math
import cv2
import numpy as np
from . import vidstab_utils


def extreme_corners(frame, transforms):
    """Calculate max drift of each frame corner caused by stabilizing transforms

    :param frame: frame from video being stabilized
    :param transforms: VidStab transforms attribute
    :return: dictionary of most extreme x and y values caused by transformations
    """
    h, w = frame.shape[:2]
    frame_corners = np.array([[0, 0],  # top left
                              [0, h - 1],  # bottom left
                              [w - 1, 0],  # top right
                              [w - 1, h - 1]],  # bottom right
                             dtype='float32')
    frame_corners = np.array([frame_corners])

    min_x = min_y = max_x = max_y = 0
    for i in range(transforms.shape[0]):
        transform = transforms[i, :]
        transform_mat = vidstab_utils.build_transformation_matrix(transform)
        transformed_frame_corners = cv2.transform(frame_corners, transform_mat)

        delta_corners = transformed_frame_corners - frame_corners

        delta_y_corners = delta_corners[0][:, 1].tolist()
        delta_x_corners = delta_corners[0][:, 0].tolist()
        min_x = min([min_x] + delta_x_corners)
        min_y = min([min_y] + delta_y_corners)
        max_x = max([max_x] + delta_x_corners)
        max_y = max([max_y] + delta_y_corners)

    return {'min_x': min_x, 'min_y': min_y, 'max_x': max_x, 'max_y': max_y}


def auto_border_start(min_corner_point, border_size):
    """Determine upper-right corner coords for auto border crop

    :param min_corner_point: extreme corner component either 'min_x' or 'min_y'
    :param border_size: min border_size determined by extreme_frame_corners in vidstab process
    :return: adjusted extreme corner for cropping
    """
    return math.floor(border_size - abs(min_corner_point))


def auto_border_length(frame_dim, extreme_corner, border_size):
    """Determine height/width auto border crop

    :param frame_dim: height/width of frame to be auto border cropped (corresponds to extreme_corner)
    :param extreme_corner: extreme corner component either 'min_x' or 'min_y' (corresponds to frame_dim)
    :param border_size: min border_size determined by extreme_frame_corners in vidstab process
    :return: adjusted extreme corner for cropping
    """
    return math.ceil(frame_dim - (border_size - extreme_corner))


def auto_border_crop(frame, extreme_frame_corners, border_size):
    """Crop frame for auto border in vidstab process

    :param frame: frame to be cropped
    :param extreme_frame_corners: extreme_frame_corners attribute of vidstab object
    :param border_size: min border_size determined by extreme_frame_corners in vidstab process
    :return: cropped frame determined by auto border process
    """
    if border_size == 0:
        return frame

    frame_h, frame_w = frame.shape[:2]

    x = auto_border_start(extreme_frame_corners['min_x'], border_size)
    y = auto_border_start(extreme_frame_corners['min_y'], border_size)

    w = auto_border_length(frame_w, extreme_frame_corners['max_x'], border_size)
    h = auto_border_length(frame_h, extreme_frame_corners['max_y'], border_size)

    return frame[y:h, x:w]


def min_auto_border_size(extreme_frame_corners):
    """Calc minimum border size to accommodate most extreme transforms

    :param extreme_frame_corners: extreme_frame_corners attribute of vidstab object
    :return: minimum border size as int
    """
    abs_extreme_corners = [abs(x) for x in extreme_frame_corners.values()]
    return math.ceil(max(abs_extreme_corners))



import warnings
import cv2
import numpy as np
import imutils
from progress.bar import IncrementalBar


def bfill_rolling_mean(arr, n=30):
    """Helper to perform trajectory smoothing

    :param arr: Numpy array of frame trajectory to be smoothed
    :param n: window size for rolling mean
    :return: smoothed input arr

    >>> arr = np.array([[1, 2, 3], [4, 5, 6]])
    >>> bfill_rolling_mean(arr, n=2)
    array([[2.5, 3.5, 4.5],
           [2.5, 3.5, 4.5]])
    """
    if arr.shape[0] < n:
        raise ValueError('arr.shape[0] cannot be less than n')
    if n == 1:
        return arr

    pre_buffer = np.zeros(3).reshape(1, 3)
    post_buffer = np.zeros(3 * n).reshape(n, 3)
    arr_cumsum = np.cumsum(np.vstack((pre_buffer, arr, post_buffer)), axis=0)
    buffer_roll_mean = (arr_cumsum[n:, :] - arr_cumsum[:-n, :]) / float(n)
    trunc_roll_mean = buffer_roll_mean[:-n, ]

    bfill_size = arr.shape[0] - trunc_roll_mean.shape[0]
    bfill = np.tile(trunc_roll_mean[0, :], (bfill_size, 1))

    return np.vstack((bfill, trunc_roll_mean))


def init_progress_bar(frame_count, max_frames, show_progress=True, gen_all=False):
    """Helper to create progress bar for stabilizing processes

    :param frame_count: input video's cv2.CAP_PROP_FRAME_COUNT
    :param max_frames: user provided max number of frames to process
    :param show_progress: user input if bar should be created
    :param gen_all: if False progress message is 'Stabilizing'; otherwise 'Generating Transforms'
    :return: a progress.bar.IncrementalBar

    >>> progress_bar = init_progress_bar(30, float('inf'))
    >>> # Stabilizing |█████████████████████████▋      | 80%
    """
    if not show_progress:
        return None

    # frame count is negative during some cv2.CAP_PROP_FRAME_COUNT failures
    bad_frame_count = frame_count <= 0
    use_max_frames = bad_frame_count or frame_count > max_frames

    if bad_frame_count and max_frames == float('inf'):
        warnings.warn('No progress bar will be shown. (Unable to grab frame count & no max_frames provided.)')
        return None

    max_bar = max_frames if use_max_frames else frame_count
    message = progress_message(gen_all)

    return IncrementalBar(message, max=max_bar, suffix='%(percent)d%%')


def progress_message(gen_all):
    """Decide progress bar message based on gen_all flag"""
    if gen_all:
        return 'Generating Transforms'
    else:
        return 'Stabilizing'


def update_progress_bar(bar, show_progress=True, finish=False):
    """helper to handle progress bar updates in vidstab process

    :param bar: progress bar to be updated
    :param show_progress: user set flag of whether or not to display progress bar
    :param finish: finish progress bar
    :return: updated progress bar
    """
    if show_progress and bar is not None:
        bar.next()

        if finish:
            bar.finish()


def playback_video(display_frame, playback_flag, delay, max_display_width=750):
    if not playback_flag:
        return False

    if display_frame.shape[1] > max_display_width:
        display_frame = imutils.resize(display_frame, width=max_display_width)

    cv2.imshow('VidStab Playback ({} frame delay if using live video;'
               ' press Q or ESC to quit)'.format(delay),
               display_frame)
    key = cv2.waitKey(1)

    if key == ord("q") or key == 27:
        return True



import numpy as np
import pytest

from vidstab.frame import Frame
from vidstab.border_utils import functional_border_sizes, crop_frame


def test_functional_border_sizes():
    assert (100, 0) == functional_border_sizes(100)
    assert (100, 110) == functional_border_sizes(-10)


def test_crop_frame():
    black_frame_image = np.zeros((100, 200, 3), dtype='uint8')
    black_frame = Frame(black_frame_image)
    border_options = {
        'neg_border_size': 10,
        'auto_border_flag': False
    }

    cropped_frame = crop_frame(black_frame, border_options)
    assert cropped_frame.image.shape == (80, 180, 3)

    border_options['neg_border_size'] = 0
    cropped_frame = crop_frame(black_frame, border_options)
    assert cropped_frame.image.shape == (100, 200, 3)

    border_options['auto_border_flag'] = True
    with pytest.raises(KeyError) as err:
        crop_frame(black_frame, border_options)

    assert 'extreme_frame_corners' in str(err.value)



import argparse
import warnings
from . import layer_overlay, VidStab


def str_int(v):
    """Handle argparse inputs to that could be str or int

    :param v: value to convert
    :return: v as int if int(v) does not raise ValueError

    >>> str_int('test')
    test
    >>> str_int(1)
    1
    """
    try:
        int_v = int(v)
        return int_v
    except ValueError:
        return v


def str_2_bool(v):
    """Convert string to bool from different possible strings

    :param v: value to convert
    :return: string converted to bool

    >>> str_2_bool('y')
    True
    >>> str_2_bool('0')
    False
    """
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def process_max_frames_arg(max_frames_arg):
    """Handle maxFrames arg in vidstab.__main__

    Convert negative values to inf

    :param max_frames_arg: maxFrames arg in vidstab.__main__
    :return: max_frames as is or inf

    >>> process_max_frames_arg(-1)
    inf
    >>> process_max_frames_arg(1)
    1
    """
    if max_frames_arg > 0:
        return max_frames_arg
    return float('inf')


def process_layer_frames_arg(layer_frames_arg):
    """Handle layerFrames arg in vidstab.__main__

    :param layer_frames_arg: layerFrames arg in vidstab.__main__
    :return: vidstab.layer_overlay if True else None

    >>> process_layer_frames_arg(False)

    """
    if layer_frames_arg:
        return layer_overlay
    return None


def process_border_size_arg(border_size_arg):
    """Handle borderSize arg in vidstab.__main__

    Convert strings that aren't 'auto' to 0

    :param border_size_arg: borderSize arg in vidstab.__main__
    :return: int or 'auto'

    >>> process_border_size_arg('a')
    0
    >>> process_border_size_arg('auto')
    'auto'
    >>> process_border_size_arg(0)
    0
    """
    if isinstance(border_size_arg, str):
        if not border_size_arg == 'auto':
            warnings.warn('Invalid borderSize provided; converting to 0.')
            border_size_arg = 0

    return border_size_arg


def cli_stabilizer(args):
    """Handle CLI vidstab processing

    :param args: result of vars(ap.parse_args()) from vidstab.__main__
    :return: None
    """

    max_frames = process_max_frames_arg(args['maxFrames'])
    border_size = process_border_size_arg(args['borderSize'])
    layer_func = process_layer_frames_arg(args['layerFrames'])

    # init stabilizer with user specified keypoint detector
    stabilizer = VidStab(kp_method=args['keyPointMethod'].upper())
    # stabilize input video and write to specified output file
    stabilizer.stabilize(input_path=args['input'],
                         output_path=args['output'],
                         smoothing_window=args['smoothWindow'],
                         max_frames=max_frames,
                         border_type=args['borderType'],
                         border_size=border_size,
                         layer_func=layer_func,
                         playback=args['playback'])



from vidstab.pop_deque import PopDeque


def test_pop_append():
    x = PopDeque([0], maxlen=2)

    assert x.pop_append(1) is None
    assert x.pop_append(2) == 0


def test_increment_append():
    x = PopDeque([0, 1, 2], maxlen=3)

    popped = x.increment_append()

    assert popped == 0
    assert list(x) == [1, 2, 3]

    popped = x.increment_append(pop_append=False)
    assert popped is None

    x.increment_append(increment=2, pop_append=False)
    assert list(x) == [3, 4, 6]



import tempfile
import argparse
import pytest
from vidstab import layer_overlay
import vidstab.main_utils as utils
from vidstab.download_videos import download_truncated_ostrich_video


def test_str_int():
    assert utils.str_int('test') == 'test'
    assert utils.str_int(1) == 1


def test_str_2_bool():
    assert utils.str_2_bool('y')
    assert not utils.str_2_bool('0')

    with pytest.raises(argparse.ArgumentTypeError) as err:
        utils.str_2_bool('foo')

    assert 'Boolean value expected.' in str(err.value)


def test_process_max_frames_arg():
    assert utils.process_max_frames_arg(-1) == float('inf')
    assert utils.process_max_frames_arg(1) == 1


def test_process_layer_frames_arg():
    assert utils.process_layer_frames_arg(False) is None
    assert utils.process_layer_frames_arg(True) == layer_overlay


def test_process_border_size_arg():
    with pytest.warns(UserWarning, match='Invalid borderSize provided; converting to 0.'):
        test_value = utils.process_border_size_arg('a')
    assert test_value == 0
    assert utils.process_border_size_arg('auto') == 'auto'
    assert utils.process_border_size_arg(100) == 100


def test_cli_stabilizer():
    """just tests that function runs"""
    tmp_dir = tempfile.TemporaryDirectory()
    truncated_ostrich_video = '{}/trunc_vid.avi'.format(tmp_dir.name)
    download_truncated_ostrich_video(truncated_ostrich_video)

    args = {
        'input': truncated_ostrich_video,
        'output': '{}/output.avi'.format(tmp_dir),
        'playback': False,
        'keyPointMethod': 'GFTT',
        'smoothWindow': 1,
        'maxFrames': -1,
        'borderType': 'black',
        'borderSize': 0,
        'layerFrames': False
    }

    utils.cli_stabilizer(args)
